#include "database.hpp"
#include <mysql/mysql.h>
#include <stdexcept>
#include <cstring>
#include <vector>
#include <ctime>

namespace {

using bind_type = decltype(MYSQL_BIND::buffer_type);
using length_type = std::remove_pointer<decltype(MYSQL_BIND::length)>::type;
using is_null_type = std::remove_pointer<decltype(MYSQL_BIND::is_null)>::type;
using error_type = std::remove_pointer<decltype(MYSQL_BIND::error)>::type;

struct mysql_param {
	length_type length;
	is_null_type is_null;

	inline mysql_param(length_type len, is_null_type null) noexcept
	: length(len)
	, is_null(null)
	{}

	inline virtual ~mysql_param() noexcept = default;
	virtual void bind(MYSQL_BIND &) noexcept = 0;
};

struct mysql_signed_param : mysql_param {

	int64_t buffer;

	inline mysql_signed_param(int64_t value) noexcept
	: mysql_param(sizeof(uint64_t), false)
	, buffer(value)
	{}

	void bind(MYSQL_BIND & b) noexcept override {
		b.buffer_type = MYSQL_TYPE_LONGLONG;
		b.buffer_length = sizeof(int64_t);
		b.is_unsigned = false;
		b.length = &length;
		b.is_null = &is_null;
		b.buffer = &buffer;
	}
};

struct mysql_unsigned_param : mysql_param {

	uint64_t buffer;

	inline mysql_unsigned_param(uint64_t value) noexcept
	: mysql_param(sizeof(uint64_t), false)
	, buffer(value)
	{}

	void bind(MYSQL_BIND & b) noexcept override {
		b.buffer_type = MYSQL_TYPE_LONGLONG;
		b.buffer_length = sizeof(uint64_t);
		b.is_unsigned = true;
		b.length = &length;
		b.is_null = &is_null;
		b.buffer = &buffer;
	}
};

struct mysql_double_param : mysql_param {

	double buffer;

	inline mysql_double_param(double value) noexcept
	: mysql_param(sizeof(double), false)
	, buffer(value)
	{}

	void bind(MYSQL_BIND & b) noexcept override {
		b.buffer_type = MYSQL_TYPE_DOUBLE;
		b.buffer_length = sizeof(double);
		b.is_unsigned = false;
		b.length = &length;
		b.is_null = &is_null;
		b.buffer = &buffer;
	}
};

struct mysql_time_param : mysql_param {

	MYSQL_TIME buffer;

	inline mysql_time_param(const datetime_t & value) noexcept
	: mysql_param(sizeof(MYSQL_TIME), false)
	, buffer(convert(value))
	{}

	void bind(MYSQL_BIND & b) noexcept override {
		b.buffer_type = MYSQL_TYPE_DATETIME;
		b.buffer_length = sizeof(MYSQL_TIME);
		b.is_unsigned = false;
		b.length = &length;
		b.is_null = &is_null;
		b.buffer = &buffer;
	}

	MYSQL_TIME convert(const datetime_t & value) noexcept {
		const auto tm = value.to_tm();
		MYSQL_TIME result = {};
		result.year = tm.tm_year + 1900;
		result.month = tm.tm_mon + 1;
		result.day = tm.tm_mday;
		result.hour = tm.tm_hour;
		result.minute = tm.tm_min;
		result.second = tm.tm_sec;
		result.second_part = value.microseconds();
		result.time_type = MYSQL_TIMESTAMP_DATETIME;
		return result;
	}
};

struct mysql_string_param : mysql_param {

	std::string buffer;

	inline mysql_string_param(const char * value)
	: mysql_param(strlen(value), false)
	, buffer(value, length)
	{}

	void bind(MYSQL_BIND & b) noexcept override {
		b.buffer_type = MYSQL_TYPE_STRING;
		b.buffer_length = buffer.size();
		b.is_unsigned = false;
		b.length = &length;
		b.is_null = &is_null;
		b.buffer = &buffer[0];
	}
};

struct mysql_null_param : mysql_param {

	inline mysql_null_param() noexcept
	: mysql_param(0, true)
	{}

	void bind(MYSQL_BIND & b) noexcept override {
		b.buffer_type = MYSQL_TYPE_NULL;
		b.buffer_length = 0;
		b.is_unsigned = false;
		b.length = &length;
		b.is_null = &is_null;
		b.buffer = nullptr;
	}
};

struct mysql_result {
	length_type length;
	is_null_type is_null;
	error_type error;

	inline mysql_result(length_type len, is_null_type null) noexcept
	: length(len)
	, is_null(null)
	, error(0)
	{}

	inline virtual ~mysql_result() noexcept = default;
	virtual void bind(MYSQL_BIND &) noexcept = 0;
	virtual void convert() {}
	virtual void fetch(size_t /*index*/, MYSQL_STMT * /*stmt*/) {}
};

struct mysql_signed_result : mysql_result {

	int64_t * ref;

	inline mysql_signed_result(int64_t & value) noexcept
	: mysql_result(sizeof(uint64_t), false)
	, ref(&value)
	{}

	void bind(MYSQL_BIND & b) noexcept override {
		b.buffer_type = MYSQL_TYPE_LONGLONG;
		b.buffer_length = sizeof(int64_t);
		b.is_unsigned = false;
		b.length = &length;
		b.is_null = &is_null;
		b.error = &error;
		b.buffer = ref;
	}

	void convert() override {
		if (is_null) {
			*ref = 0;
		}
	}
};

struct mysql_unsigned_result : mysql_result {

	uint64_t * ref;

	inline mysql_unsigned_result(uint64_t & value) noexcept
	: mysql_result(sizeof(uint64_t), false)
	, ref(&value)
	{}

	void bind(MYSQL_BIND & b) noexcept override {
		b.buffer_type = MYSQL_TYPE_LONGLONG;
		b.buffer_length = sizeof(uint64_t);
		b.is_unsigned = true;
		b.length = &length;
		b.is_null = &is_null;
		b.error = &error;
		b.buffer = ref;
	}

	void convert() override {
		if (is_null) {
			*ref = 0;
		}
	}
};

struct mysql_double_result : mysql_result {

	double * ref;

	inline mysql_double_result(double & value) noexcept
	: mysql_result(sizeof(double), false)
	, ref(&value)
	{}

	void bind(MYSQL_BIND & b) noexcept override {
		b.buffer_type = MYSQL_TYPE_DOUBLE;
		b.buffer_length = sizeof(double);
		b.is_unsigned = false;
		b.length = &length;
		b.is_null = &is_null;
		b.error = &error;
		b.buffer = ref;
	}

	void convert() override {
		if (is_null) {
			*ref = 0;
		}
	}
};

struct mysql_time_result : mysql_result {

	MYSQL_TIME buffer;
	datetime_t * ref;

	inline mysql_time_result(datetime_t & value) noexcept
	: mysql_result(sizeof(MYSQL_TIME), false)
	, ref(&value)
	{}

	void bind(MYSQL_BIND & b) noexcept override {
		b.buffer_type = MYSQL_TYPE_DATETIME;
		b.buffer_length = sizeof(MYSQL_TIME);
		b.is_unsigned = false;
		b.length = &length;
		b.is_null = &is_null;
		b.error = &error;
		b.buffer = &buffer;
	}

	void convert() override {
		if (is_null) {
			ref->set(0, 0);
		} else {
			struct tm tm = {};
			tm.tm_year = buffer.year - 1900;
			tm.tm_mon = buffer.month - 1;
			tm.tm_mday = buffer.day;
			tm.tm_hour = buffer.hour;
			tm.tm_min = buffer.minute;
			tm.tm_sec = buffer.second;
			ref->set(tm, buffer.second_part);
		}
	}
};

struct mysql_string_result : mysql_result {

	std::string buffer;
	const char ** ref;

	inline mysql_string_result(const char * & value) noexcept
	: mysql_result(0, false)
	, ref(&value)
	{}

	void bind(MYSQL_BIND & b) noexcept override {
		b.buffer_type = MYSQL_TYPE_STRING;
		b.buffer_length = buffer.size();
		b.is_unsigned = false;
		b.length = &length;
		b.is_null = &is_null;
		b.error = &error;
		b.buffer = &buffer[0];
	}

	void convert() override {
		if (is_null) {
			*ref = "";
		} else {
			*ref = buffer.data();
		}
	}

	void fetch(size_t index, MYSQL_STMT * stmt) override {
		buffer.resize(length);
		MYSQL_BIND b;
		this->bind(b);
		if (mysql_stmt_fetch_column(stmt, &b, index, 0) != 0) {
			throw std::runtime_error(mysql_stmt_error(stmt));
		}
		*ref = buffer.data();
	}
};

struct mysql_null_result : mysql_result {

	inline mysql_null_result() noexcept
	: mysql_result(0, true)
	{}

	void bind(MYSQL_BIND & b) noexcept override {
		b.buffer_type = MYSQL_TYPE_NULL;
		b.buffer_length = 0;
		b.is_unsigned = false;
		b.length = &length;
		b.is_null = &is_null;
		b.error = &error;
		b.buffer = nullptr;
	}
};

struct mysql_statement : db_statement {

	MYSQL_STMT * stmt;
	bool inputs_changed, outputs_changed;
	std::vector<std::unique_ptr<mysql_param>> inputs;
	std::vector<std::unique_ptr<mysql_result>> outputs;

	mysql_statement(MYSQL * mysql)
	: stmt(mysql_stmt_init(mysql))
	, inputs_changed(true)
	, outputs_changed(true)
	{
		if (!stmt) {
			throw std::runtime_error(mysql_error(mysql));
		}
	}

	mysql_statement(MYSQL * mysql, const std::string & query)
	: mysql_statement(mysql)
	{
		if (mysql_stmt_prepare(stmt, query.c_str(), query.size()) != 0) {
			throw std::runtime_error(mysql_stmt_error(stmt));
		}
		inputs.resize(mysql_stmt_param_count(stmt));
		for (size_t i = 0; i < inputs.size(); ++i) {
			inputs[0].reset(new mysql_null_param());
		}
	}

	~mysql_statement() noexcept override {
		mysql_stmt_close(stmt);
	}

	void bind_param(size_t index, int64_t value) override {
		inputs[index].reset(new mysql_signed_param(value));
		inputs_changed = true;
	}

	void bind_param(size_t index, uint64_t value) override {
		inputs[index].reset(new mysql_unsigned_param(value));
		inputs_changed = true;
	}

	void bind_param(size_t index, double value) override {
		inputs[index].reset(new mysql_double_param(value));
		inputs_changed = true;
	}

	void bind_param(size_t index, const datetime_t & value) override {
		inputs[index].reset(new mysql_time_param(value));
		inputs_changed = true;
	}

	void bind_param(size_t index, const char * value) override {
		inputs[index].reset(new mysql_string_param(value));
		inputs_changed = true;
	}

	void bind_param(size_t index, decltype(nullptr)) override {
		inputs[index].reset(new mysql_null_param());
		inputs_changed = true;
	}

	void execute() override {
		if (inputs_changed) {
			std::vector<MYSQL_BIND> binds(inputs.size());
			for (size_t i = 0; i < inputs.size(); ++i) {
				inputs[i]->bind(binds[i]);
			}
			if (mysql_stmt_bind_param(stmt, binds.data()) != 0) {
				throw std::runtime_error(mysql_stmt_error(stmt));
			}
			inputs_changed = false;
		}
		if (mysql_stmt_execute(stmt) != 0) {
			throw std::runtime_error(mysql_stmt_error(stmt));
		}
		outputs.resize(mysql_stmt_field_count(stmt));
		for (size_t i = 0; i < outputs.size(); ++i) {
			outputs[i].reset(new mysql_null_result());
		}
	}

	void bind_result(size_t index, int64_t & value) override {
		outputs[index].reset(new mysql_signed_result(value));
		outputs_changed = true;
	}

	void bind_result(size_t index, uint64_t & value) override {
		outputs[index].reset(new mysql_unsigned_result(value));
		outputs_changed = true;
	}

	void bind_result(size_t index, double & value) override {
		outputs[index].reset(new mysql_double_result(value));
		outputs_changed = true;
	}

	void bind_result(size_t index, datetime_t & value) override {
		outputs[index].reset(new mysql_time_result(value));
		outputs_changed = true;
	}

	void bind_result(size_t index, const char * & value) override {
		outputs[index].reset(new mysql_string_result(value));
		outputs_changed = true;
	}

	bool fetch() override {
		if (outputs_changed) {
			std::vector<MYSQL_BIND> binds(outputs.size());
			for (size_t i = 0; i < outputs.size(); ++i) {
				outputs[i]->bind(binds[i]);
			}
			if (mysql_stmt_bind_result(stmt, binds.data()) != 0) {
				throw std::runtime_error(mysql_stmt_error(stmt));
			}
			outputs_changed = false;
		}
		const auto result = mysql_stmt_fetch(stmt);
		if (result == 0) {
			for (const auto & output : outputs) {
				output->convert();
			}
			return true;
		} else if (result == 1) {
			throw std::runtime_error(mysql_stmt_error(stmt));
		} else if (result == MYSQL_DATA_TRUNCATED) {
			for (size_t i = 0; i < outputs.size(); ++i) {
				if (outputs[i]->error) {
					outputs[i]->fetch(i, stmt);
				} else {
					outputs[i]->convert();
				}
			}
			outputs_changed = true;
			return true;
		}
		return false;
	}
};

struct mysql_database : database {

	MYSQL mysql;

	mysql_database() {
		if (!mysql_init(&mysql)) {
			throw std::runtime_error("mysql_init failed");
		}
	}

	mysql_database(
		const std::string & hostname,
		const std::string & username,
		const std::string & password,
		const std::string & name)
	: mysql_database()
	{
		mysql_options(&mysql, MYSQL_INIT_COMMAND,
			"SET NAMES utf8 COLLATE utf8_unicode_ci, time_zone = '+0:00'");
		const auto result = mysql_real_connect(
			&mysql, hostname.c_str(),
			username.c_str(), password.c_str(),
			name.c_str(), 0, nullptr, 0);
		if (!result) {
			mysql_close(&mysql);
			throw std::runtime_error(mysql_error(&mysql));
		}
	}

	~mysql_database() noexcept override {
		mysql_close(&mysql);
	}

	std::unique_ptr<db_statement> prepare(const std::string & qs) override {
		return std::unique_ptr<db_statement>(new mysql_statement(&mysql, qs));
	}
};

} // anonymous

std::unique_ptr<database> database::connect(
	proto,
	const std::string & hostname,
	const std::string & username,
	const std::string & password,
	const std::string & name
) {
	return std::unique_ptr<database>(new mysql_database(hostname, username, password, name));
}
