#ifndef __DATABASE_HPP__
#define __DATABASE_HPP__

#include "polyfill.hpp"
#include "deduce.hpp"
#include "datetime.hpp"
#include <string>
#include <memory>
#include <utility>
#include <initializer_list>
#include <type_traits>
#include <tuple>

struct db_statement {
	inline virtual ~db_statement() noexcept = default;

	virtual void bind_param(size_t index, int64_t value) = 0;
	virtual void bind_param(size_t index, uint64_t value) = 0;
	virtual void bind_param(size_t index, double value) = 0;
	virtual void bind_param(size_t index, const datetime_t & value) = 0;
	virtual void bind_param(size_t index, const char * value) = 0;
	void bind_param(size_t index, const std::string & value);
	virtual void bind_param(size_t index, decltype(nullptr)) = 0;
	virtual void bind_result(size_t index, int64_t & value) = 0;
	virtual void bind_result(size_t index, uint64_t & value) = 0;
	virtual void bind_result(size_t index, double & value) = 0;
	virtual void bind_result(size_t index, datetime_t & value) = 0;
	virtual void bind_result(size_t index, const char * & value) = 0;
	virtual void execute() = 0;
	virtual bool fetch() = 0;

	void bind_params(size_t index);

	template<typename T>
	void bind_params(size_t index, T && value);

	template<typename T, typename ...Ts>
	void bind_params(size_t index, T && value, Ts &&... values);

	template<typename T>
	void bind_results(size_t index, T & value);

	template<typename T, typename ...Ts>
	void bind_results(size_t index, T & value, Ts &... values);

	template<typename Tuple, size_t ...I>
	void bind_results(Tuple &, index_sequence<I...>);

	template<typename Function>
	void fetch(Function);
};

struct database {

	enum proto {
		mysql,
	};

	inline virtual ~database() noexcept = default;

	static std::unique_ptr<database> connect(
		proto protocol,
		const std::string & hostname,
		const std::string & username,
		const std::string & password,
		const std::string & name);
	virtual std::unique_ptr<db_statement> prepare(const std::string & qs) = 0;

	template<typename ...Ts>
	std::unique_ptr<db_statement> query(const std::string & qs, Ts &&... params);
};

template<typename ...Ts>
inline std::unique_ptr<db_statement> database::query(const std::string & qs, Ts &&... params) {
	auto stmt = prepare(qs);
	stmt->bind_params(0, std::forward<Ts>(params)...);
	stmt->execute();
	return stmt;
}

inline void db_statement::bind_param(size_t index, const std::string & value) {
	bind_param(index, value.c_str());
}

inline void db_statement::bind_params(size_t) {}

template<typename T>
inline void db_statement::bind_params(size_t index, T && value) {
	bind_param(index, std::forward<T>(value));
}

template<typename T, typename ...Ts>
inline void db_statement::bind_params(size_t index, T && value, Ts &&... values) {
	bind_param(index, std::forward<T>(value));
	bind_params(index + 1, std::forward<Ts>(values)...);
}

template<typename T>
inline void db_statement::bind_results(size_t index, T & value) {
	bind_result(index, value);
}

template<typename T, typename ...Ts>
inline void db_statement::bind_results(size_t index, T & value, Ts &... values) {
	bind_result(index, value);
	bind_results(index + 1, values...);
}

template<typename Tuple, size_t ...I>
inline void db_statement::bind_results(Tuple & tuple, index_sequence<I...>) {
	bind_results(0, std::get<I>(tuple)...);
}

template<typename Function>
inline void db_statement::fetch(Function fn) {
	using tuple_type = tuple_type_t<Function>;
	constexpr auto num_fields = std::tuple_size<tuple_type>::value;
	tuple_type tuple;
	bind_results(tuple, make_index_sequence<num_fields>{});
	while (fetch()) {
		apply(fn, tuple);
	}
}

#endif
