#ifndef __DATE_TIME_HPP__
#define __DATE_TIME_HPP__

#include "interval.hpp"
#include <sys/time.h>
#include <cstdint>
#include <ctime>
#include <string>

class datetime_t {
public:

	datetime_t() noexcept;
	datetime_t(time_t seconds, suseconds_t microseconds = 0) noexcept;
	datetime_t(const struct tm &, suseconds_t microseconds = 0);

	time_t seconds() const noexcept;
	int64_t milliseconds() const noexcept;
	suseconds_t microseconds() const noexcept;

	datetime_t & set(time_t seconds, suseconds_t microseconds = 0) noexcept;
	datetime_t & set(const struct tm &, suseconds_t microseconds = 0);

	struct tm to_tm() const;
	std::string format(const char * format = "%Y-%m-%d %H:%M:%S") const;
	bool zero() const noexcept;

	static datetime_t now();

	bool operator==(const datetime_t &) const noexcept;
	bool operator!=(const datetime_t &) const noexcept;
	bool operator<(const datetime_t &) const noexcept;
	bool operator>(const datetime_t &) const noexcept;
	bool operator<=(const datetime_t &) const noexcept;
	bool operator>=(const datetime_t &) const noexcept;

	timeval operator-(const datetime_t &) const;
	datetime_t operator+(const interval_t &) const;

protected:
	timeval value;
};

inline datetime_t::datetime_t() noexcept : value({}) {}

inline datetime_t::datetime_t(time_t sec, suseconds_t usec) noexcept {
	set(sec, usec);
}

inline datetime_t::datetime_t(const struct tm & tm, suseconds_t usec) {
	set(tm, usec);
}

inline time_t datetime_t::seconds() const noexcept {
	return value.tv_sec;
}

inline int64_t datetime_t::milliseconds() const noexcept {
	return (value.tv_sec * 1000) + (value.tv_usec / 1000);
}

inline suseconds_t datetime_t::microseconds() const noexcept {
	return value.tv_usec;
}

inline datetime_t & datetime_t::set(time_t sec, suseconds_t usec) noexcept {
	value.tv_sec = sec;
	value.tv_usec = usec;
	return *this;
}

inline bool datetime_t::operator==(const datetime_t & other) const noexcept {
	return value.tv_sec == other.value.tv_sec &&
		value.tv_usec == other.value.tv_usec;
}

inline bool datetime_t::operator!=(const datetime_t & other) const noexcept {
	return value.tv_sec != other.value.tv_sec ||
		value.tv_usec != other.value.tv_usec;
}

inline bool datetime_t::operator<(const datetime_t & other) const noexcept {
	return value.tv_sec < other.value.tv_sec ||
		(value.tv_sec == other.value.tv_sec && value.tv_usec < other.value.tv_usec);
}

inline bool datetime_t::operator>(const datetime_t & other) const noexcept {
	return value.tv_sec > other.value.tv_sec ||
		(value.tv_sec == other.value.tv_sec && value.tv_usec > other.value.tv_usec);
}

inline bool datetime_t::operator<=(const datetime_t & other) const noexcept {
	return value.tv_sec < other.value.tv_sec ||
		(value.tv_sec == other.value.tv_sec && value.tv_usec <= other.value.tv_usec);
}

inline bool datetime_t::operator>=(const datetime_t & other) const noexcept {
	return value.tv_sec > other.value.tv_sec ||
		(value.tv_sec == other.value.tv_sec && value.tv_usec >= other.value.tv_usec);
}

inline bool datetime_t::zero() const noexcept {
	return value.tv_sec == 0 && value.tv_usec == 0;
}

inline timeval datetime_t::operator-(const datetime_t & other) const {
	timeval result;
	timersub(&value, &other.value, &result);
	return result;
}

#endif
