#include "scheduler.hpp"
#include "database.hpp"
#include "utils.hpp"
#include "signal_handler.hpp"
#include "socket_handler.hpp"
#include <system_error>
#include <cstring>
#include <cinttypes>
#include <vector>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/types.h>
#include <sys/signalfd.h>
#include <sys/epoll.h>
#include <signal.h>

namespace {

const auto socket_filename = "/var/run/cpbx-scheduler/cpbx-scheduler.sock";

void create_epollfd() {
	scheduler.epollfd = epoll_create1(EPOLL_CLOEXEC);
	if (scheduler.epollfd < 0) {
		throw std::system_error(errno, std::system_category());
	}
}

void create_signalfd() {
	sigset_t sigmask;
	if (sigemptyset(&sigmask) != 0) {
		throw std::system_error(errno, std::system_category());
	}
	if (sigaddset(&sigmask, SIGCHLD) != 0) {
		throw std::system_error(errno, std::system_category());
	}
	if (sigprocmask(SIG_BLOCK, &sigmask, nullptr) != 0) {
		throw std::system_error(errno, std::system_category());
	}
	scheduler.signalfd = ::signalfd(-1, &sigmask, SFD_NONBLOCK | SFD_CLOEXEC);
	if (scheduler.signalfd < 0) {
		throw std::system_error(errno, std::system_category());
	}
}

void create_socketfd() {
	scheduler.socketfd = socket(AF_UNIX, SOCK_STREAM | SOCK_NONBLOCK | SOCK_CLOEXEC, 0);
	if (scheduler.socketfd < 0) {
		throw std::system_error(errno, std::system_category());
	}
	sockaddr_un addr = {};
	addr.sun_family = AF_UNIX;
	strncpy(addr.sun_path, socket_filename, sizeof(addr.sun_path) - 1);
	if (::bind(scheduler.socketfd, reinterpret_cast<sockaddr *>(&addr), sizeof(addr)) != 0) {
		throw std::system_error(errno, std::system_category());
	}
	if (listen(scheduler.socketfd, SOMAXCONN) != 0) {
		throw std::system_error(errno, std::system_category());
	}
}

void process_event(const epoll_event & event) {
	const auto event_handler = static_cast<event_handler_t *>(event.data.ptr);
	if (scheduler.event_handlers.count(event_handler) > 0) {
		if (event.events & (EPOLLIN | EPOLLOUT)) {
			event_handler->trigger();
		}
		if (event.events & (EPOLLERR | EPOLLHUP | EPOLLRDHUP)) {
			scheduler.unregister_handler(event_handler);
		}
	}
}

std::vector<task_t *> runnable_tasks() {
	std::vector<task_t *> list;
	for (auto & kv : scheduler.tasks) {
		auto & task = kv.second;
		const auto finished = task.oneshot && task.starting < task.completed;
		if (task.enabled && !task.running() && !finished) {
			list.emplace_back(&task);
		}
	}
	return list;
}

int calculate_timeout() {
	int timeout = -1;
	const auto tasks = runnable_tasks();
	if (!tasks.empty()) {
		const auto now = datetime_t::now();
		auto earliest = tasks[0]->next_run;
		for (const auto task : tasks) {
			if (task->next_run < now) {
				earliest = now;
			} else if (task->next_run < earliest) {
				earliest = task->next_run;
			}
		}
		timeout = earliest.milliseconds() - now.milliseconds();
	}
	return std::min(std::max(timeout, 1000), 60000);
}

void execute_tasks() {
	const auto now = datetime_t::now();
	const auto tasks = runnable_tasks();
	for (auto task : tasks) {
		if (task->next_run <= now) {
			try {
				task->execute();
			} catch (const std::exception & e) {
				log("Task %" PRIu64 " (%s) failed to start: %s",
					task->id, task->script.c_str(), e.what());
				task->reschedule(now);
				const auto scheduled = task->next_run.format();
				log("Task %" PRIu64 " (%s) scheduled for %s UTC",
					task->id, task->script.c_str(), scheduled.c_str());
			}
		}
	}
}

void print_schedule() {
	for (const auto & kv : scheduler.tasks) {
		const auto & task = kv.second;
		if (!task.enabled) {
			log("Task %" PRIu64 " (%s) disabled", task.id, task.script.c_str());
		} else if (task.running()) {
			log("Task %" PRIu64 " (%s) running", task.id, task.script.c_str());
		} else if (task.oneshot && task.starting < task.completed) {
			log("Task %" PRIu64 " (%s) completed", task.id, task.script.c_str());
		} else {
			const auto scheduled = task.next_run.format();
			log("Task %" PRIu64 " (%s) scheduled for %s UTC",
				task.id, task.script.c_str(), scheduled.c_str());
		}
	}
}

void wait_for_events() {
	epoll_event events[16];
	const auto num_events = sizeof(events) / sizeof(*events);
	const auto timeout_ms = calculate_timeout();
	const auto result = epoll_wait(scheduler.epollfd, events, num_events, timeout_ms);
	if (result > 0) {
		for (int i = 0; i < result; ++i) {
			process_event(events[i]);
		}
	} else if (result == 0) {
		execute_tasks();
	} else if (errno != EINTR) {
		throw std::system_error(errno, std::system_category());
	}
}

struct cpbx_config_t {
	std::string hostname;
	std::string username;
	std::string password;
	std::string database;
};

cpbx_config_t parse_cpbx_config(const std::string & input) {
	const auto first_newline = input.find_first_of('\n');
	if (first_newline == input.npos) {
		throw std::runtime_error("CPBX configuration file parse error");
	}
	const auto second_newline = input.find_first_of('\n', first_newline + 1);
	if (second_newline == input.npos) {
		throw std::runtime_error("CPBX configuration file parse error");
	}
	const auto third_newline = input.find_first_of('\n', second_newline + 1);
	if (third_newline == input.npos) {
		throw std::runtime_error("CPBX configuration file parse error");
	}
	return cpbx_config_t{
		input.substr(0, first_newline),
		input.substr(first_newline + 1, second_newline - first_newline - 1),
		input.substr(second_newline + 1, third_newline - second_newline - 1),
		input.substr(third_newline + 1),
	};
}

cpbx_config_t read_cpbx_config() {
	// Let PHP do all the parsing
	const auto command = "php -r '"
		"require_once(\"/etc/ombutel/ombutel.conf\");"
		"printf(\"%s\n%s\n%s\n%s\","
			"config::database_host,"
			"config::database_username,"
			"config::database_password,"
			"config::database_name);"
		"'";
	std::unique_ptr<FILE, decltype(&pclose)> fp(popen(command, "r"), pclose);
	if (!fp) {
		throw std::system_error(errno, std::system_category());
	}
	std::vector<char> buffer(4096);
	size_t written = 0;
	while (true) {
		const auto amount = buffer.size() - written;
		if (amount == 0) {
			// buffer full, resize
			buffer.resize(buffer.size() + 4096);
			continue;
		}
		const auto dst = &buffer[written];
		const auto res = fread(dst, 1, amount, fp.get());
		written += res;
		if (res < amount) {
			if (feof(fp.get())) {
				break; // done
			} else if (ferror(fp.get())) {
				throw std::runtime_error("Cannot parse CPBX configuration file");
			}
		}
	}
	return parse_cpbx_config(std::string(buffer.data(), written));
}

std::unique_ptr<database> open_database() {
	const auto config = read_cpbx_config();
	return database::connect(
		database::mysql,
		config.hostname,
		config.username,
		config.password,
		config.database
	);
}

} // anonymous

scheduler_t scheduler;

void scheduler_t::start() {
	load_tasks();
	create_epollfd();
	create_signalfd();
	create_socketfd();
	register_handler(signalfd, new signal_handler_t(signalfd));
	register_handler(socketfd, new socket_handler_t(socketfd));
	while (true) {
		wait_for_events();
	}
}

void scheduler_t::register_handler(int fd, event_handler_t * event_handler) {
	epoll_event event = {};
	event.data.ptr = event_handler;
	event.events = EPOLLIN | EPOLLOUT | EPOLLRDHUP | EPOLLET;
	if (epoll_ctl(epollfd, EPOLL_CTL_ADD, fd, &event) != 0) {
		throw std::system_error(errno, std::system_category());
	}
	event_handlers.emplace(event_handler);
}

void scheduler_t::unregister_handler(event_handler_t * event_handler) {
	if (event_handlers.count(event_handler) > 0) {
		if (epoll_ctl(epollfd, EPOLL_CTL_DEL, event_handler->fd, nullptr) != 0) {
			throw std::system_error(errno, std::system_category());
		}
		delete event_handler;
		event_handlers.erase(event_handler);
	}
}

void scheduler_t::load_tasks() {
	try {
		std::map<uint64_t, task_t> new_tasks;
		const auto db = open_database();
		const auto stmt = db->query(
			"select "
				"`task_id`,"
				"`script`,"
				"`parameters`,"
				"`starting`,"
				"`completed`,"
				"`interval`,"
				"`interval_unit`,"
				"`enabled`,"
				"`exact`"
			"from `ombu_scheduled_tasks`");
		stmt->fetch([ &new_tasks ] (
			uint64_t id,
			const char * script,
			const char * parameters,
			datetime_t starting,
			datetime_t completed,
			uint64_t interval_amount,
			const char * interval_unit,
			const char * enabled,
			const char * exact
		) {
			if (strlen(script) > 0) {
				const auto interval = interval_t::parse(interval_amount, interval_unit);
				new_tasks.emplace(id, task_t::parse(id, script, parameters,
					starting, completed, interval, enabled, exact));
			}
		});
		tasks.swap(new_tasks);
		print_schedule();
	} catch (const std::exception & e) {
		log("Unable to load tasks from database: %s", e.what());
	}
}
