#include "signal_handler.hpp"
#include "database.hpp"
#include "scheduler.hpp"
#include "utils.hpp"
#include <system_error>
#include <cinttypes>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/signalfd.h>

namespace {

std::unique_ptr<database> open_database() {
	return database::connect(database::mysql, "localhost", "root", "", "ombutel");
}

void update(uint64_t task_id, uint64_t exitcode, const datetime_t & now) {
	try {
		const auto db = open_database();
		db->query(
			"update `ombu_scheduled_tasks` "
			"set `completed` = ?, `exitcode` = ? "
			"where `task_id` = ?",
			now, exitcode, task_id);
	} catch (const std::exception & e) {
		log("Unable to update task %" PRIu64 ": %s", task_id, e.what());
	}
}

void terminated(pid_t pid, int exitcode) {
	const auto it1 = scheduler.processes.find(pid);
	if (it1 != scheduler.processes.end()) {
		const auto it2 = scheduler.tasks.find(it1->second);
		if (it2 != scheduler.tasks.end()) {
			auto & task = it2->second;
			log("Task %" PRIu64 " (%s) terminated with code %d",
				task.id, task.script.c_str(), exitcode);
			const auto now = datetime_t::now();
			update(task.id, exitcode, now);
			task.completed = now;
			const auto finished = task.oneshot && task.starting < task.completed;
			if (task.enabled && !finished) {
				task.reschedule(now);
				const auto scheduled = task.next_run.format();
				log("Task %" PRIu64 " (%s) scheduled for %s UTC",
					task.id, task.script.c_str(), scheduled.c_str());
			}
		}
		scheduler.processes.erase(it1);
	}
}

void wait_for_process(pid_t pid) {
	int wstatus;
	const auto result = waitpid(pid, &wstatus, WUNTRACED | WCONTINUED);
	if (result < 0 && errno != ECHILD) {
		throw std::system_error(errno, std::system_category());
	} else if (WIFEXITED(wstatus)) {
		return terminated(pid, WEXITSTATUS(wstatus));
	} else if (WIFSIGNALED(wstatus)) {
		return terminated(pid, WTERMSIG(wstatus));
	} else if (WIFSTOPPED(wstatus)) {
		return terminated(pid, WSTOPSIG(wstatus));
	}
}

bool read_signal() {
	signalfd_siginfo si;
	const auto result = read(scheduler.signalfd, &si, sizeof(si));
	if (sizeof(si) == result) {
		if (si.ssi_signo == SIGCHLD) {
			wait_for_process(si.ssi_pid);
		}
		return true;
	} else if (result >= 0) {
		return false; // short read
	} else if (errno == EINTR) {
		return true;
	} else if (errno == EAGAIN || errno == EWOULDBLOCK) {
		return false;
	} else {
		throw std::system_error(errno, std::system_category());
	}
}

} // anonymous

void signal_handler_t::trigger() {
	try {
		while (read_signal()) {}
	} catch (const std::exception & e) {
		log("Signal error: %s", e.what());
	}
}
