#include "socket_handler.hpp"
#include "scheduler.hpp"
#include "utils.hpp"
#include <system_error>
#include <utility>
#include <sys/socket.h>
#include <unistd.h>

namespace {

void execute(uint64_t task_id) {
	const auto it = scheduler.tasks.find(task_id);
	if (it != scheduler.tasks.end()) {
		const auto & task = it->second;
		task.execute();
	}
}

struct client_handler_t : event_handler_t {

	inline client_handler_t(int _fd) noexcept : event_handler_t(_fd) {}

	~client_handler_t() noexcept override {
		close(fd);
	}

	void trigger() override {
		try {
			while (read_task_id(fd)) {}
		} catch (const std::exception & e) {
			log("Client error: %s", e.what());
			scheduler.unregister_handler(this);
		}
	}

	bool read_task_id(int fd) {
		uint64_t task_id;
		const auto result = read(fd, &task_id, sizeof(task_id));
		if (sizeof(task_id) == result) {
			if (task_id == 0) {
				scheduler.load_tasks();
			} else {
				execute(task_id);
			}
			return true; // ok, next value
		} else if (result == 0) {
			// remote hung up
			scheduler.unregister_handler(this);
			return false;
		} else if (result > 0) {
			throw std::runtime_error("short read");
		} else if (errno == EINTR) {
			return true; // try again
		} else if (errno == EAGAIN || errno == EWOULDBLOCK) {
			return false; // try again later
		} else {
			throw std::system_error(errno, std::system_category());
		}
	}
};

bool accept_socket() {
	const auto fd = accept4(scheduler.socketfd, nullptr, nullptr, SOCK_NONBLOCK | SOCK_CLOEXEC);
	if (fd >= 0) {
		scheduler.register_handler(fd, new client_handler_t(fd));
		return true;
	} else if (errno == EINTR) {
		return true;
	} else if (errno == EAGAIN || errno == EWOULDBLOCK) {
		return false;
	} else {
		throw std::system_error(errno, std::system_category());
	}
}

} // anonymous

void socket_handler_t::trigger() {
	try {
		while (accept_socket()) {}
	} catch (const std::exception & e) {
		log("Socket error: %s", e.what());
	}
}
