#include "task.hpp"
#include "scheduler.hpp"
#include "utils.hpp"
#include <stdexcept>
#include <system_error>
#include <cstring>
#include <cinttypes>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <unistd.h>

const auto task_scripts_dir = "/usr/share/ombutel/task-scripts/";

namespace {

bool file_exists(const std::string & filename) {
	struct stat sb;
	if (stat(filename.c_str(), &sb) == 0) {
		return true;
	} else if (errno != ENOENT) {
		throw std::system_error(errno, std::system_category());
	}
	return false;
}

void exec(std::string command) {
	if (setpriority(PRIO_PROCESS, 0, 19) != 0) {
		throw std::system_error(errno, std::system_category());
	}
	std::string copies[] = { "sh", "-c", task_scripts_dir+command };
	char * args[4] = {
		&copies[0][0],
		&copies[1][0],
		&copies[2][0],
		nullptr,
	};
	if (execvp("sh", args) < 0) {
		throw std::system_error(errno, std::system_category());
	}
}

double random() {
	return double(rand()) / double(RAND_MAX);
}

bool is_leap_year(int year) {
	if ((year % 4) != 0) {
		return false;
	} else if ((year % 100) != 0) {
		return true;
	} else if ((year % 400) != 0) {
		return false;
	}
	return true;
}

datetime_t next_start_minutes(
	const datetime_t & starting,
	const datetime_t & now,
	const interval_t & interval
) {
	const auto interval_sec = interval.amount * 60;
	const auto diff = now.seconds() - starting.seconds();
	const auto remainder = diff % interval_sec;
	const auto starting_tm = starting.to_tm();
	auto tm = now.to_tm();
	if (remainder == 0) {
		tm.tm_sec = starting_tm.tm_sec;
	} else {
		tm.tm_sec += interval_sec - remainder;
	}
	return datetime_t(tm, now.microseconds());
}

datetime_t next_start_hours(
	const datetime_t & starting,
	const datetime_t & now,
	const interval_t & interval
) {
	const auto interval_sec = interval.amount * 3600;
	const auto remainder = (now - starting).tv_sec % interval_sec;
	const auto starting_tm = starting.to_tm();
	auto tm = now.to_tm();
	if (remainder == 0) {
		tm.tm_sec = starting_tm.tm_sec;
		tm.tm_min = starting_tm.tm_min;
	} else {
		tm.tm_sec += interval_sec - remainder;
	}
	return datetime_t(tm, now.microseconds());
}

datetime_t next_start_days(
	const datetime_t & starting,
	const datetime_t & now,
	const interval_t & interval
) {
	const auto interval_sec = interval.amount * 86400;
	const auto remainder = (now - starting).tv_sec % interval_sec;
	const auto starting_tm = starting.to_tm();
	auto tm = now.to_tm();
	if (remainder == 0) {
		tm.tm_sec = starting_tm.tm_sec;
		tm.tm_min = starting_tm.tm_min;
		tm.tm_hour = starting_tm.tm_hour;
	} else {
		tm.tm_sec += interval_sec - remainder;
	}
	return datetime_t(tm, now.microseconds());
}

datetime_t next_start_weeks(
	const datetime_t & starting,
	const datetime_t & now,
	const interval_t & interval
) {
	const auto interval_sec = interval.amount * 604800;
	const auto remainder = (now - starting).tv_sec % interval_sec;
	const auto starting_tm = starting.to_tm();
	auto tm = now.to_tm();
	if (remainder == 0) {
		tm.tm_sec = starting_tm.tm_sec;
		tm.tm_min = starting_tm.tm_min;
		tm.tm_hour = starting_tm.tm_hour;
	} else {
		tm.tm_sec += interval_sec - remainder;
	}
	return datetime_t(tm, now.microseconds());
}

datetime_t next_start_months(
	const datetime_t & starting,
	const datetime_t & now,
	const interval_t & interval
) {
	static const int days_per_month[2][12] = {
		{ 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
		{ 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
	};
	auto tm = starting.to_tm();
	const auto now_tm = now.to_tm();
	const auto diff = ((now_tm.tm_year - tm.tm_year) * 12) +
		(now_tm.tm_mon - tm.tm_mon);
	const auto multiplier = 1 + (diff / interval.amount);
	tm.tm_year += (multiplier * interval.amount) / 12;
	tm.tm_mon = tm.tm_mon + ((multiplier * interval.amount) % 12);
	if (tm.tm_mon >= 12) {
		tm.tm_mon = tm.tm_mon % 12;
		tm.tm_year += 1;
	}
	const auto month_days = days_per_month[is_leap_year(tm.tm_year)][tm.tm_mon];
	if (tm.tm_mday > month_days) {
		tm.tm_mday = month_days;
	}
	return datetime_t(tm, now.microseconds());
}

datetime_t calculate_next_start(
	const datetime_t & starting,
	const datetime_t & now,
	const interval_t & interval,
	bool enabled,
	bool oneshot
) {
	if (enabled && !oneshot && starting < now) {
		switch (interval.unit) {
			case interval_t::minutes: return next_start_minutes(starting, now, interval);
			case interval_t::hours: return next_start_hours(starting, now, interval);
			case interval_t::days: return next_start_days(starting, now, interval);
			case interval_t::weeks: return next_start_weeks(starting, now, interval);
			case interval_t::months: return next_start_months(starting, now, interval);
			default: return starting;
		}
	}
	return starting;
}

datetime_t next_run_fixed(
	const datetime_t & next_start
) {
	const interval_t fudge(interval_t::minutes, 30.0 * random());
	return next_start + fudge;
}

datetime_t next_run_seconds(
	const datetime_t & next_start,
	const interval_t & interval
) {
	const auto seconds = double(interval.amount);
	const interval_t fudge(interval_t::seconds, seconds * random());
	return next_start + fudge;
}

datetime_t next_run_minutes(
	const datetime_t & next_start,
	const interval_t & interval
) {
	if (interval.amount > 30) {
		return next_run_fixed(next_start);
	}
	const auto seconds = double(interval.amount) * 60.0;
	const interval_t fudge(interval_t::seconds, seconds * random());
	return next_start + fudge;
}

datetime_t calculate_next_run(
	const datetime_t & next_start,
	const datetime_t & completed,
	const interval_t & interval,
	bool enabled,
	bool is_exact
) {
	if (enabled && !is_exact && next_start > completed) {
		switch (interval.unit) {
			case interval_t::seconds: return next_run_seconds(next_start, interval);
			case interval_t::minutes: return next_run_minutes(next_start, interval);
			case interval_t::hours: return next_run_fixed(next_start);
			case interval_t::days: return next_run_fixed(next_start);
			case interval_t::weeks: return next_run_fixed(next_start);
			case interval_t::months: return next_run_fixed(next_start);
			default: return next_start;
		}
	}
	return next_start;
}

} // anonymous

bool task_t::running() const {
	for (const auto & kv : scheduler.processes) {
		if (kv.second == id) {
			return true;
		}
	}
	return false;
}

void task_t::execute() const {
	if (running()) {
		log("Task %" PRIu64 " (%s) already running", id, script.c_str());
	} else {
		auto command = script;
		if (!parameters.empty()) {
			command += " " + parameters;
		}
		if (file_exists(task_scripts_dir+script)) {
			const auto pid = fork();
			if (pid == 0) {
				exec(command);
			} else if (pid > 0) {
				log("Task %" PRIu64 " (%s) running, pid %ld",
					id, script.c_str(), pid);
				scheduler.processes.emplace(pid, id);
			} else {
				throw std::system_error(errno, std::system_category());
			}
		} else {
			const auto message = format("Executable for task %" PRIu64 " (%s) not found",
				id, script.c_str());
			throw std::runtime_error(message);
		}
	}
}

void task_t::reschedule(const datetime_t & now) {
	next_start = calculate_next_start(starting, now, interval, enabled, oneshot);
	next_run = calculate_next_run(next_start, completed, interval, enabled, exact);
}

task_t task_t::parse(
	uint64_t id,
	const char * script,
	const char * parameters,
	const datetime_t & starting,
	const datetime_t & completed,
	const interval_t & interval,
	const char * enabled,
	const char * exact
) {
	const auto is_enabled = strcmp(enabled, "yes") == 0;
	const auto is_exact = strcmp(exact, "yes") == 0;
	const auto oneshot = interval.amount == 0;
	const auto next_start = calculate_next_start(starting, datetime_t::now(), interval, is_enabled, oneshot);
	const auto next_run = calculate_next_run(next_start, completed, interval, is_enabled, is_exact);
	return task_t {
		id,
		script,
		parameters,
		interval,
		starting,
		completed,
		next_start,
		next_run,
		is_enabled,
		is_exact,
		oneshot,
	};
}
